// -*- C++ -*-
/*!
 * @file  GTKPickview.cpp
 * @brief 3D Viewer
 * @date $Date$
 *
 * $Id$
 */

#include "GTKPickview.h"

// Module specification
// <rtc-template block="module_spec">
static const char* gtkpickview_spec[] =
  {
    "implementation_id", "GTKPickview",
    "type_name",         "GTKPickview",
    "description",       "3D Viewer",
    "version",           "1.0.0",
    "vendor",            "Hiroaki Yaguchi, JSK",
    "category",          "Viewer",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "0",
    "language",          "C++",
    "lang_type",         "compile",
    "conf.default.PointCloudColor", "0",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
GTKPickview::GTKPickview(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_Map2DIn("Map2D", m_Map2D),
    m_MapConfigIn("MapConfig", m_MapConfig),
    m_Box3DIn("Box3D", m_Box3D),
    m_Points3DIn("Points3D", m_Points3D),
    m_ColorImageIn("ColorImage", m_ColorImage),
    m_DepthImageIn("DepthImage", m_DepthImage)

    // </rtc-template>
{
  thread_running=false;
  glo=NULL;
  pthread_mutex_init(&thread_lock,NULL);
  startGTKThread(this);
}

/*!
 * @brief destructor
 */
GTKPickview::~GTKPickview()
{
  if(thread_running) stopGTKThread(this);
  pthread_mutex_destroy(&thread_lock);
  if(glo!=NULL) delete glo;
  delete gview;
}


void GTKPickview::setPoints(vector<float> &pts){
  pthread_mutex_lock(&thread_lock);
  vec_points.assign(pts.begin(),pts.end());
  pthread_mutex_unlock(&thread_lock);
}
void GTKPickview::setPointsColor(vector<unsigned char> &col){
  pthread_mutex_lock(&thread_lock);
  vec_pointscolor.assign(col.begin(),col.end());
  pthread_mutex_unlock(&thread_lock);
}
void GTKPickview::setMap(int width,int height,vector<float> &map){
  pthread_mutex_lock(&thread_lock);
  map_width=width;
  map_height=height;
  vec_map.assign(map.begin(),map.end());
  pthread_mutex_unlock(&thread_lock);
}
void GTKPickview::setMapConfig(float xs,float ys,Pose2D &pose){
  pthread_mutex_lock(&thread_lock);
  map_xscale=xs;
  map_yscale=ys;
  map_pose=pose;
  pthread_mutex_unlock(&thread_lock);
}
void GTKPickview::setBox(vector<float> &box){
  pthread_mutex_lock(&thread_lock);
  vec_box.assign(box.begin(),box.end());
  pthread_mutex_unlock(&thread_lock);
}
void GTKPickview::getPoints(vector<float> &pts){
  pthread_mutex_lock(&thread_lock);
  pts.assign(vec_points.begin(),vec_points.end());
  pthread_mutex_unlock(&thread_lock);
}
void GTKPickview::getPointsColor(vector<unsigned char> &col){
  pthread_mutex_lock(&thread_lock);
  col.assign(vec_pointscolor.begin(),vec_pointscolor.end());
  pthread_mutex_unlock(&thread_lock);
}
void GTKPickview::getMap(vector<float> &map){
  pthread_mutex_lock(&thread_lock);
  map.assign(vec_map.begin(),vec_map.end());
  pthread_mutex_unlock(&thread_lock);
}
void GTKPickview::getBox(vector<float> &box){
  pthread_mutex_lock(&thread_lock);
  box.assign(vec_box.begin(),vec_box.end());
  pthread_mutex_unlock(&thread_lock);
}

void GTKPickview::createGLObjects(){
  gl_grid=new GLGrid();
  gl_grid->setGrid(20,100.0);
  gl_grid->setTranslate(1000.0, 0.0, 0.0);

  gl_box=new GLBox(100.0,100.0,100.0);
  gl_box->setColor(1.0,1.0,0.0,1.0);
  gl_box->setTranslate(1000.0, 0.0, 0.0);
  //gl_box->setGLList();
  gl_box->useList(false);
  gl_box->setVisible(false);

  gl_voxel=new GLVoxel(100.0,100.0,100.0);
  gl_voxel->setTranslate(1000.0, 0.0, 0.0);

  gl_points=new GLFeaturePoints();
  gl_points->setTranslate(1000.0, 0.0, 0.0);
}



RTC::ReturnCode_t GTKPickview::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("Map2D", m_Map2DIn);
  addInPort("MapConfig", m_MapConfigIn);
  addInPort("Box3D", m_Box3DIn);
  addInPort("Points3D", m_Points3DIn);
  addInPort("ColorImage", m_ColorImageIn);
  addInPort("DepthImage", m_DepthImageIn);
  
  // Set OutPort buffer
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>
  bindParameter("PointCloudColor", m_PointCloudColor, "0");

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t GTKPickview::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t GTKPickview::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t GTKPickview::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t GTKPickview::onActivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t GTKPickview::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t GTKPickview::onExecute(RTC::UniqueId ec_id)
{
  int len;
  vector<float> buf;
  vector<unsigned char> colorbuf;

  if(m_Map2DIn.isNew()){
    //cout << " [onExecute]: map recieved" << endl;
    m_Map2DIn.read();
    len=m_Map2D.cells.length();
    buf.resize(len);
    //memcpy(&buf[0],&m_Map2D.data[0],sizeof(float)*len);
    for(int i=0;i<len;i++)
      buf[i]=(float)m_Map2D.cells[i]/100.0;
    //setMap(buf);
    setMap(m_Map2D.width,m_Map2D.height,buf);
  }

  if(m_MapConfigIn.isNew()){
    //cout << " [onExecute]: map config recieved" << endl;
    m_MapConfigIn.read();
    setMapConfig((float)m_MapConfig.xScale,(float)m_MapConfig.yScale,m_MapConfig.origin);
  }

  if(m_Box3DIn.isNew()){
    //cout << " [onExecute]: box recieved" << endl;
    m_Box3DIn.read();
    len=m_Box3D.data.length();
    buf.resize(len);
    memcpy(&buf[0],&m_Box3D.data[0],sizeof(float)*len);
    setBox(buf);
  }

  if(m_Points3DIn.isNew()){
    //cout << " [onExecute]: points recieved" << endl;
    m_Points3DIn.read();
    //len=m_Points3D.data.length();
    len=m_Points3D.points.length();
    buf.resize(len*3);
    //memcpy(&buf[0],&m_Points3D.data[0],sizeof(float)*len);
    for(int i=0;i<len;i++){
      buf[i*3]=m_Points3D.points[i].point.x;
      buf[i*3+1]=m_Points3D.points[i].point.y;
      buf[i*3+2]=m_Points3D.points[i].point.z;
    }
    setPoints(buf);
    if(m_PointCloudColor == 1){
      colorbuf.resize(len*3);
      for(int i=0;i<len;i++){
	colorbuf[i*3]=(unsigned char)(m_Points3D.points[i].colour.r * 255);
	colorbuf[i*3+1]=(unsigned char)(m_Points3D.points[i].colour.g * 255);
	colorbuf[i*3+2]=(unsigned char)(m_Points3D.points[i].colour.b * 255);
      }
      setPointsColor(colorbuf);
    }
  }

  if(m_DepthImageIn.isNew()){
    //cout << " [onExecute]: depth recieved" << endl;
    m_DepthImageIn.read();

    if(640 == m_DepthImage.width &&
       480 == m_DepthImage.height &&
       11 == m_DepthImage.bpp &&
       (640*480*2) == m_DepthImage.pixels.length()){
      buf.resize(640*480*3);
      int xidx,idx;
      int xpidx,pidx;

      int maxDepth=2048; //=2^11
      int stepDepth=maxDepth/32;
      vector <unsigned int> depthHistogram(32);
      for(int i=0;i<depthHistogram.size();i++)
	depthHistogram[i]=0;

      float minDistance=-100;
      float scaleFactor=0.0021;
      for(int i=0;i<480;i++){
	xidx=i*640;
	for(int j=0;j<640;j++){
	  idx=xidx+j;
	  unsigned short depth=
	    m_DepthImage.pixels[idx*2]+
	    (m_DepthImage.pixels[idx*2+1]<<8);
	  depthHistogram[depth/stepDepth]++;

	  float zz=1000.0*0.1236*tan((float)(depth)/2842.5+1.1863);

	  if(zz>0.0){
	    // convert x,y,z -> -y,-z,x 
	    buf[idx*3+1]=
	      -(float)((j-320)*(zz+minDistance)*scaleFactor)*4.0/3.0;
	    buf[idx*3+2]=1000.0
	      -(float)((i-240)*(zz+minDistance)*scaleFactor);
	    buf[idx*3]=(float)zz;
	  }else{
	    buf[idx*3]=buf[idx*3+1]=buf[idx*3+2]=0.0;
	  }
	}
      }
      setPoints(buf);
      for(int i=0;i<depthHistogram.size();i++)
	cout << depthHistogram[i] << " ";
      cout << endl;
    }
  }

  if(m_ColorImageIn.isNew()){
    m_ColorImageIn.read();
    if(640 == m_ColorImage.width &&
       480 == m_ColorImage.height &&
       24 == m_ColorImage.bpp &&
       (640*480*3) == m_ColorImage.pixels.length()){
      colorbuf.resize(640*480*3);
      for(int i=0;i<640*480*3;i++)
	colorbuf[i]=m_ColorImage.pixels[i];
    }
    setPointsColor(colorbuf);
  }

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t GTKPickview::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t GTKPickview::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t GTKPickview::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t GTKPickview::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t GTKPickview::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


// for pthread
void startGTKThread(GTKPickview *gp){
  pthread_t tid;
  
  fprintf(stderr,"start thread...\n");

  if(!gp->threadRunning()){
    fprintf(stderr,"  standing by.\n");
    gp->standbyThread();
    fprintf(stderr,"  creating thread...\n");
    pthread_create(&tid,NULL,GTKThreadRoutine,(void *)gp);
    fprintf(stderr,"  complete...\n");
    gp->setThreadID(tid);
  }

  fprintf(stderr,"done.\n");
}
void stopGTKThread(GTKPickview *gp){
  pthread_t tid=gp->getThreadID();
  if(gp->threadRunning()){
    gp->destroyThread();
    pthread_join(tid,NULL);
  }
}
void *GTKThreadRoutine(void *arg){
  GTKPickview *gp=(GTKPickview*)arg;

  // v hontouni korede iino?
  int argc=0;
  char **argv=NULL;

  unsigned int timeout_id=0;

  gtk_set_locale();
  gtk_init(&argc,&argv);
  gdk_rgb_init();
  gtk_gl_init(&argc,&argv);

  GtkView *gview=gp->createGtkView();
  gview->activatePickView();
  gview->show();

  //gp->setGLObject((GLObject *)new GLGrid());
  //gp->setGLObject((GLObject *)new GLBox(100.0,100.0,100.0));
  //gview->setGLObject(gp->getGLObject());
  gp->createGLObjects();
  gview->setGLObject((GLObject *)gp->getGLGrid());
  gview->addGLObject((GLObject *)gp->getGLBox());
  gview->addGLObject((GLObject *)gp->getGLFeaturePoints());
  gview->addGLObject((GLObject *)gp->getGLVoxel());

  //float fv[]={0,300,800,-PI/2.0,0,0};
  float fv[]={0,800,800,0,-PI/2.0,PI/2.0};
  gview->setMarkPose(fv);

  timeout_id=gtk_timeout_add(33, (GtkFunction)polling_timer_event, arg);

  cout << "gtk_main();" << endl;
  gtk_main();
  cout << "gtk_main_quit();" << endl;
}

static gboolean polling_timer_event(gpointer p){
  GTKPickview *gp=(GTKPickview*)p;
  GtkView *gview=gp->getGtkView();
  pthread_mutex_t *mutex=gp->mutex();
  if(!gp->threadRunning()){
    gtk_main_quit();
    return TRUE;
  }

  // update drawing commands
  vector<float> buf;
  int mwidth,mheight;
  float mx,my;
  float mpx,mpy,mth;
  float pi=PI;

  mx=gp->getMapXScale();
  my=gp->getMapYScale();
  mpx=gp->getMapPosX();
  mpy=gp->getMapPosY();
  mth=gp->getMapRot();
  gp->getGLVoxel()->setVoxelSize(mx,my,100.0);
  gp->getGLVoxel()->setTranslate(1000.0+mpx,mpy,0.0);
  gp->getGLVoxel()->setRotate(0.0,0.0,mth);

  gp->getMap(buf);
  mwidth=gp->getMapWidth();
  mheight=gp->getMapHeight();
  gp->getGLVoxel()->setMap(mwidth,mheight,buf);

  gp->getPoints(buf);
  gp->getGLFeaturePoints()->setPoints(buf);

  float rr,gg,bb;
  float dmin=500.0;
  float dmax=3000.0;
  float dstep=(dmax-dmin)/4.0;
  int hcidx;
  vector<float> colors;
  vector<unsigned char> ucolors;
  if(buf.size()>=3){
    gp->getPointsColor(ucolors);
    if(buf.size()!=ucolors.size()){
      colors.resize(buf.size());
      gp->getGLFeaturePoints()->useColorList(true);
      for(int i;i<buf.size();i+=3){
	if(buf[i]<dmin){
	  rr=1.0; gg=0.0; bb=0.0;
	}else if(buf[i]>dmax){
	  rr=0.0; gg=0.0; bb=1.0;
	}else{
	  hcidx=(int)((buf[i]-dmin)/dstep);
	  switch(hcidx){
	  case 0:
	    rr=1.0;
	    gg=(buf[i]-dmin)/dstep;
	    bb=0.0;
	    break;
	  case 1:
	    rr=2.0-(buf[i]-dmin)/dstep;
	    gg=1.0;
	    bb=0.0;
	    break;
	  case 2:
	    rr=0.0;
	    gg=1.0;
	    bb=(buf[i]-dmin)/dstep-2.0;
	    break;
	  case 3:
	    rr=0.0;
	    gg=4.0-(buf[i]-dmin)/dstep;
	    bb=1.0;
	    break;
	  default:
	    rr=0.0; gg=0.0; bb=0.0;
	  }
	}
	colors[i]=rr;
	colors[i+1]=gg;
	colors[i+2]=bb;
      }
    }else{
      colors.resize(ucolors.size());
      gp->getGLFeaturePoints()->useColorList(true);
      for(int i=0;i<ucolors.size();i++)
	colors[i]=(float)(ucolors[i])/255.0;
    }
  }
  gp->getGLFeaturePoints()->setColorList(colors);

  gp->getBox(buf);
  if(buf.size()>=9){
    /*
    cout << " box: ";
    for(int i=0;i<9;i++) cout << buf[i] << " ";
    cout << endl;
    */
    gp->getGLBox()->setVisible(true);
    // box is (tx ty tz rx ry rz depth width height)
    gp->getGLBox()->setTranslate(buf[0]+1000.0,buf[1],buf[2]);
    gp->getGLBox()->setRotate(180.0*buf[3]/pi,180.0*buf[4]/pi,180.0*buf[5]/pi);
    gp->getGLBox()->setSize(buf[6],buf[7],buf[8]);
  }else{
    gp->getGLBox()->setVisible(false);
  }

  gview->redraw();

  return TRUE;
}

extern "C"
{
 
  void GTKPickviewInit(RTC::Manager* manager)
  {
    coil::Properties profile(gtkpickview_spec);
    manager->registerFactory(profile,
                             RTC::Create<GTKPickview>,
                             RTC::Delete<GTKPickview>);
  }
  
};


